/*-
 * Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2000..2014 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: fancy.c,v 1.2 2014/09/29 20:13:04 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  3

#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sysexits.h>
#include <sys/uio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <time.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include "deleted.h"

#define POSIT 13

static const char *maskn[] =
{ "mtime", "ctime", "btime", "lock"
, "atime", "point", "owner", "group"
, "mode" , NULL   , NULL   , NULL
, NULL   , NULL   , NULL   , "NOrm"
};

void
debug_date(FILE *f, time_t time) {
    struct tm *t;

    t = localtime(&time);
    fprintf( f, "%04d-%02d-%02d %02d:%02d:%02d UTC"
           , 1900 + t->tm_year, t->tm_mon, t->tm_mday
           , t->tm_hour, t->tm_min, t->tm_sec
           );
}

static int
debug_time(FILE *f, time_t time) {
    time_t tim = time;
    unsigned t;
    int ex = 0;

    for (t = 0; tim >= 31556952; t++) tim -= 31556952;
    if  (t) ex += fprintf(f, "%uy", t);
    for (t = 0; tim >= 604800; t++) tim -= 604800;
    if  (t) ex += fprintf(f, "%uw", t);
    for (t = 0; tim >= 86400; t++) tim -= 86400;
    if  (t) ex += fprintf(f, "%ud", t);
    for (t = 0; tim >= 3600; t++) tim -= 3600;
    if  (t) ex += fprintf(f, "%uh", t);
    for (t = 0; tim >= 60; t++) tim -= 60;
    if  (t) ex += fprintf(f, "%um", t);
    if  (tim || !time) ex += fprintf(f, "%us", (unsigned)tim);
    return(ex);
}

int
query(deleted_pid *parm) {
    int pidfd, i, ex = EX_OK;
    FILE *f = stdout;
    pid_t pid;
    char *o;

    parm->conf = NULL;
    if  (0 > (pidfd = open(parm->pidname, O_RDONLY, 0600))) {
        ifBLIN_QW0("Control file not found");
        EXOUT(EX_IOERR);
    }
    fprintf(f, "Control file found");
    if  (!flock(pidfd, LOCK_EX | LOCK_NB)) {                          /* If no other deleted running */
        fprintf(f, ". Daemon not found.\n");
        if  (!!flock(pidfd, LOCK_SH | LOCK_NB)) {
            ifBLIN_QW0("flock shared after exclusive");
            EXOUT(EX_UNAVAILABLE);
        }
    } else if (EWOULDBLOCK != errno) {
        ifBLIN_QW0("EX flock of control file not possible");
    } else if (!flock(pidfd, LOCK_SH | LOCK_NB)) {
        fprintf(f, ".\n");
    } else if (EWOULDBLOCK != errno) {
        ifBLIN_QW0("SH flock of control file not possible");
    } else {
        fprintf(f, ", it is blocked.\n");
    }
    parm->conf = mmap(NULL, sizeof(deleted_pid), PROT_READ, MAP_PRIVATE, pidfd, (off_t)0);
    if  (MAP_FAILED == (void*)parm->conf) {
        ifBLIN_QW0("mmap");
        EXOUT(EX_UNAVAILABLE);
    }
    if  (parm->conf->id != parm->id) {
        ifBLIN_QX0("%s belongs to program %016X <> %016X", parm->pidname, parm->conf->id, parm->id);
        EXOUT(EX_DATAERR);
    }
    if  (parm->conf->version != parm->version) {
        ifBLIN_QX0( "%s belongs to version %u <> %u"
                  , parm->pidname, parm->conf->version, parm->version
                  );
        EXOUT(EX_DATAERR);
    }
    if  (parm->conf->dpss != parm->dpss) {
        ifBLIN_QX0("%s belongs to arch %u <> %u", parm->pidname, parm->conf->dpss, parm->dpss);
        EXOUT(EX_DATAERR);
    }
    errno = 0;
    pid = strtoul(parm->conf->cpid, &o, 10);
    if  (!!errno) {
        ifBLIN_QW0("strtoul PID %s", parm->conf->cpid);
        EXOUT(EX_IOERR);
    } else if ('\n' != *o) {
        ifBLIN_QX0("Invalidated PID %s", parm->conf->cpid);
        EXOUT(EX_DATAERR);
    } else if (pid != parm->conf->pid) {
        ifBLIN_QX0("!!! (cpid == %d) != (pid == %d) !!!", pid, parm->conf->pid);
    } else {
        fprintf(f, "pid == %d", pid);
        if  (!kill(pid, 0)) {
            fprintf(f, " running\n");
        } else if (EINVAL == errno) {
            fprintf(f, " not permited\n");
        } else {
            fprintf(f, " not running\n");
    }   }
    fprintf(f, "%08X Nice = %d\n", parm->conf->flags, parm->conf->prio);
    fprintf(f, "Wake up at ");
    debug_date(stdout, parm->conf->wake_up);
    fprintf(f, "\nDirectory ttl=");
    i = debug_time(f, parm->conf->dmask.ttl);
    for (; i < POSIT; ++i) fprintf(f, " ");
    fprintf(f, " mask=");
    i = blin_64(parm->conf->dmask.mask, maskn, 16, f);
    fprintf(f, "\nFile      ttl=");
    i = debug_time(f, parm->conf->fmask.ttl);
    for (; i < POSIT; ++i) fprintf(f, " ");
    fprintf(f, " mask=");
    i = blin_64(parm->conf->fmask.mask, maskn, 16, f);
    fprintf(f, "\nSymLink   ttl=");
    i = debug_time(f, parm->conf->lmask.ttl);
    for (; i < POSIT; ++i) fprintf(f, " ");
    fprintf(f, " mask=");
    i = blin_64(parm->conf->lmask.mask, maskn, 16, f);
    fprintf(f, "\n");
out:
    return(ex);
}
